/*
 * ATtiny841_MAX487_032_Slave3.cpp
 * Created: 03.01.2018 15:42:10
 *
 * C : ATtiny841 mit internen 8MHz Oszillator
 *
 * USART0/1 mit usart Lib von Peter Fleury, mit Ringbuffer usw.
 *
 * Interrupt-Vector Namen bersicht:
 * C:\Program Files (x86)\Atmel\Studio\7.0\toolchain\avr8\avr8-gnu-toolchain\avr\include\avr\iotn841.h
 *
 *  Projekt -> meinProjekt Properties -> Konfiguration > alle Konfigurationen
 *  Projekt -> meinProjekt Properties -> C/C++ Compiler > Symbols > Defined symbols:
 *  eintragen:   F_CPU=8000000UL
 *
 *  Projekt -> meinProjekt Properties -> C++ Compiler > Miscellaneous:
 *  eintragen:   -std=c++11		(wegen constexpr)
 *
 * neuer Code zur Servopulserzeugung
 * USART0 hatte Receive ISR enabled Probleme durch Timer 1 Compare A ISR
 * https://www.mikrocontroller.net/topic/441958
 * Umbau handle_Function_UART0_Pins() auf switch-case
 * Verwendung des UART0 TXB80 Bits eingebaut
*/


#include <util/delay.h>
#include <avr/io.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/interrupt.h>
#include <util/crc16.h>
#include <util/atomic.h>   // fr cli() und sei() mit SREG Sicherung
#include "pinDefi.h"		   // eigene I/O Definitionen
#include "usart.h"			// rechts > 2. > hinzufgen > vorhandenes Element
#include "timer.h"			// alles was mit Timern zu tun hat ausgelagert

#define NOP __asm__ __volatile__ ("nop")

#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))   // setzt das angegebene Bit auf 1
#endif
#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))  // setzt (lscht) das angegebene Bit auf 0
#endif


// UART Berechnungen des Wertes fr das Baudratenregister aus Taktrate und gewnschter Baudrate
#define UART_BAUD_RATE	250000UL    // gewnschte Baudrate
#define UBRR_VAL ((F_CPU+UART_BAUD_RATE*8)/(UART_BAUD_RATE*16)-1)	// sauber runden
#define BAUD_REAL (F_CPU/(16*(UBRR_VAL+1)))							// reale Baudrate
#define BAUD_ERROR ((BAUD_REAL*1000)/UART_BAUD_RATE)				// Fehler in Promille, 1000 = kein Fehler.
#if ((BAUD_ERROR<995) || (BAUD_ERROR>1005))
  #error Systematischer Fehler der Baudrate groesser 0,5% und damit zu hoch! 
#endif


// Definitionen und Initialisierungen
extern volatile uint16_t UART0_RxBuf[UART_RX_BUFFER_SIZE];
extern volatile uint16_t UART0_TxBuf[UART_TX_BUFFER_SIZE];
extern volatile uint8_t UART0_TxHead;
extern volatile uint8_t UART0_TxTail;

uint32_t last_MasterRequestTime;

uint8_t org_OSCCAL;
typedef enum {UART, SYNC, WAITSYNC, RUN, CALC, REPEAT, FINISH, ERROR} state_t;    // Steuerzustnde
bool volatile state_measurement_complete = false;

uint16_t countCalibration;
uint16_t countErrorsUART0;		// uart0 Fehlerzhler
uint32_t uart_Timeout;

const uint8_t STX = '\xCC';		// Startkennzeichen der Nachricht
const uint8_t ETX = '\xEE';		// Endekennzeichen der Nachricht
const uint8_t ESC = '\xDD';		// 
const uint8_t MasterAddr = 0;	   // #0 Adresse vom Master
const uint8_t myAddr = 3;		   // #3 meine Slave Adresse

// definiere das Datenprotokoll
union Message
{
	struct
	{
		uint8_t toAddr;
		uint8_t fromAddr;
		uint8_t cmd;
		int32_t data;
		uint16_t crc;
	};
	uint8_t asArray[9];			// Summe aller struct Datentypen, fr Zugriff ber Index
} empfDaten, sendDaten;			// zwei gleiche union Buffer anlegen

const uint8_t length_Message = sizeof(Message);

uint32_t global_ms;

// *** Funktion Deklarationen *** //
// *** USART *** //
void Txd0Rxd0_normalPortPins();
void uart0_disable();
bool read_Ringbuffer_0();
void sendStruct(uint8_t *structPtr, uint8_t length);
void handle_Serial_0_to_Serial_0();
void handle_Function_UART0_Pins();
void send_Nachricht();
uint16_t calc_CRC16 (uint8_t feld[], uint8_t length);
// *** allgemeine *** //
void set_PCINT2();
void del_PCINT2();
bool Calibration_Calculation();


int main(void)
{  				
	Pin_TxD0_OUT;	// TxD0 Ausgang
	Pin_RxD0_IN;	// RxD0 Eingang
	Pin_RxD0_PULL;	// RxD0 Pullup aktiv
	
	MAX487_TXRX_OUT;	// PA0 Ausgang, MAX487 /RE_DE 
	MAX487_empfangen;	// MAX487 /RE_DE - umschalten auf empfangen
	
	Led1_OUT;	// Pin auf Ausgang konfigurieren
	Led2_OUT;
	Led3_OUT;
	Led4_OUT;
	Led5_OUT;
	Led6_OUT;
	// Led7_OUT;   Quarz
	// Led8_OUT;   Quarz
		
	preSet_Timer0();			// millis, seconds
	preSet_Timer1();
	preSet_Timer2();			// Timer zum einmessen/kalibrieren
			   
	sendDaten.toAddr = MasterAddr;   
	sendDaten.fromAddr = myAddr;
	sendDaten.data = 123;
   org_OSCCAL = OSCCAL0;		// Debug
   
   while (1)	// Hauptprogramm
   {   					
	   handle_Function_UART0_Pins();
		//global_ms = millis();
	}
}


/* *** Funktionen *** */
void handle_Function_UART0_Pins()
{
  //static state_t state_UART_MODE = SYNC;			// Startbedingung
  static state_t state_UART_MODE = FINISH;	   	// mit externen 8MHz Quarz
   
  switch (state_UART_MODE) {
		
	case UART:	// UART0 wird benutzt
	            handle_Serial_0_to_Serial_0();
               if (countErrorsUART0 > 50000) {  // erstmal ne wilde Hausnummer
		            state_UART_MODE = ERROR;
	            }
			break;		// minimal 7,7s ... maximal 18,2s
	
	case SYNC:	// schaltet UART0 Funktionalitt ab
				   uart0_disable();
				   Txd0Rxd0_normalPortPins();
				   state_UART_MODE = WAITSYNC;
			break;
		
	case WAITSYNC:	// OSCCAL Calibration ready, dauert 85s
				   memset((void *)UART0_TxBuf, '\0', sizeof(UART0_TxBuf));	// Messwerte lschen
				   memset((void *)UART0_RxBuf, '\0', sizeof(UART0_RxBuf)); 
				   stop_Timer0();
				   stop_Timer1();
				   set_PCINT2();						   // ISR(PCINT0_vect) scharf schalten
				   run_Timer2();						
				   state_UART_MODE = RUN;
			break;
	
	case RUN:	// Takteinmesszyklus luft mittels ISR(PCINT0_vect)
				   if (state_measurement_complete == true) {
					   state_UART_MODE = CALC;
				   }
			break;
				
	case CALC:	// OSCCAL neu berechnen
				   if (Calibration_Calculation() == true) {
					    state_UART_MODE = FINISH;
				   }
				   else {
					    state_UART_MODE = REPEAT;
                   state_measurement_complete = false; // Reset fr ISR(PCINT0_vect) einmessen
				   }
			break;
				
	case REPEAT:	// Kalibrierung wiederholen
				   set_PCINT2();
				   countCalibration++;
				   state_UART_MODE = WAITSYNC;
			break;
	
	case FINISH:	// schaltet UART0 Funktion ein
               stop_Timer2();	
				   del_PCINT2();
				   run_Timer0();
				   run_Timer1();
				   memset((void *)UART0_TxBuf, '\0', sizeof(UART0_TxBuf));  // Messwerte lschen
				   memset((void *)UART0_RxBuf, '\0', sizeof(UART0_RxBuf));
				   uart0_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) );
				   state_UART_MODE = UART;
			break;
	
		
	case ERROR:	// Fehlerbehandlung noch nicht komplett
				   // uart0_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) );
				   // oder 
				   // state_UART_MODE = SYNC;
				   state_UART_MODE = UART;
				   // countErrorsUART0 = 0;
			break;
				
	default:	break;
  }
}


void set_PCINT2()
{
	GIMSK = (1<<PCIE0);         // PCINT 7:0 Gruppe Trigger aktiv
	PCMSK0 = (1<<PCINT2);		// PCINT2, Pin Einzelauswahl
}


void del_PCINT2()
{
	GIMSK = 0;      // PCINT 7:0 Gruppe deaktivieren
	PCMSK0 = 0;		// PCINT2 deaktivieren
}


bool Calibration_Calculation ()			// dauert komplett max. 357s
{
	bool state_Finish = false;
	const uint16_t Target = 128;		   // gltig fr Baudrate 62500 und 8MHz
	const uint16_t min = Target-15;		// falsche Messwerte filtern
	const uint16_t max = Target+15;
	uint16_t summe = 0;
	uint16_t diff = 0;
	uint16_t count_fail = 0;
	uint16_t Ticks = 0;
		
	// TCNT2 auswerten ... 59 Differenzen bilden
	ATOMIC_BLOCK (ATOMIC_RESTORESTATE) {				// dauert 270s
		for (uint8_t i=0; i<29; i++) {					// 0...28           
			diff = (UART0_RxBuf[i+1] - UART0_RxBuf[i]);	// 29 Differenzen
			if (min < diff && diff < max) {				// filtert falsche Werte
				summe = summe + diff;
			}
			else {
				count_fail++;				// Anzahl ungltiger Messwerte zhlen
			}
		}

		diff = (UART0_TxBuf[0] - UART0_RxBuf[29]);		// 30. Differenz
		if (min < diff && diff < max) {					// filtert falsche Werte
			summe = summe + diff;
		}
		else {
			count_fail++;					// Anzahl ungltiger Messwerte zhlen
		}

		for (uint8_t i=0; i<29; i++) {					// 0...28
			diff = (UART0_TxBuf[i+1] - UART0_TxBuf[i]);	// weitere 29 Differenzen
			if (min < diff && diff < max) {				// filtert falsche Werte
				summe = summe + diff;
			}
			else {
				count_fail++;				// Anzahl ungltiger Messwerte zhlen
			}
		}		
	}
	
	Ticks = (1.0*summe/(59-count_fail))+0.5;  // Durchschnitt berechnen und runden		
						
	if (Ticks > Target) {
		OSCCAL0--;
		state_Finish = false;
	}
		
	if (Ticks < Target) {
		OSCCAL0++;
		state_Finish = false;
			
	}
	    
	if (Ticks == Target) {			// Ticks auf Ziel ?			
		state_Finish = true;		   // Kalibrierung beenden
	}
		
	if (count_fail > 6) {			// zu viele ungltige Messwerte
		state_Finish = false;
	}	
	
	return state_Finish;	
}


ISR(PCINT0_vect)	// Interrupt Handler fr PCINTs, dauert 3,5s
{   				
	uint16_t data = TCNT2;
	static uint8_t index = 0;		// index = cycle
	
	if ( index < 30 ) {				// 0...29
		UART0_RxBuf[index] = data;	// speichert TNCT2 im Rx Buffer
	}
	else {
		UART0_TxBuf[index-30] = data;	// 30...59 im Tx Buffer
	}
   
	index++;
	
	if (index > 59) {				// nach 60 Pegelwechsel Auswertung
		del_PCINT2();				// ISR temporr sperren
		state_measurement_complete = true;
		index = 0;
	}
}


void Txd0Rxd0_normalPortPins()
{
	Pin_TxD0_OUT;		// TxD0 Ausgang
	Pin_RxD0_IN;		// RxD0 Eingang
	Pin_RxD0_PULL;		// RxD0 Pullup aktiv
}


void uart0_disable()
{
	UBRR0H = 0;
	UBRR0L = 0;
	UCSR0B = 0;
	UCSR0C = 0;
}


bool read_Ringbuffer_0()
{   
	//  mit Fehlerbehandlung und nichts tun, no data, wird die Funktion nach 0,5 s verlassen 
	// ohne Fehlerbehandlung und nichts tun, wird die Funktion nach 10,46 s verlassen 
	
	static uint8_t index = 0;
	static bool state_Read = false;
	static bool state_Complete = false;
	static bool state_Bypass = false;
		
	uint16_t c = uart0_getc();				// nchstes Zeichen vom Ringbuffer holen
	
	// UART Lib Fehlercodes	(dezimal)
	//      4096			  2048				  1024				 512                   256
	// UART_FRAME_ERROR, UART_OVERRUN_ERROR, UART_PARITY_ERROR, UART_BUFFER_OVERFLOW, UART_NO_DATA
	
	if ( c & UART_NO_DATA) {	
		return false;						// Abbruch nach 500ns, es gibt nichts zum lesen
	}
	
	if ( c > UART_NO_DATA) {				// irgendein UART Error
		countErrorsUART0++;					// uart0 Fehlerzhler
		index = 0;
		state_Read = false;
		state_Complete = false;
		state_Bypass = false;
		return false;						// es gibt nichts zum lesen
	}
	
	// ------------------------------------------------------------------
	c &= 0xFF;								// High Byte nullen vor Weiterverarbeitung
		
	if (state_Bypass == false) {            // Byte speziell auswerten
		if (c == ESC) {						// ESC Kennung ?
			state_Bypass = true;
			return false;
		}
		else if (c == ETX) {				// ENDE Kennung ?
			state_Read = false;
			state_Complete = true;
		}
		else if (c == STX) {				// START Kennung ?
			index = 0;
			state_Read = true;
			state_Complete = false;
			return false;
		}
	}	

	if (state_Read == true && (index < length_Message)) {	// Bytes einsortieren
		empfDaten.asArray[index++] = (uint8_t)c;
		state_Bypass = false;
	}
	
	if (state_Complete == true && index >= length_Message ) {		// bertragungsende
		state_Complete = false;
		index = 0;
		if ( calc_CRC16(empfDaten.asArray, length_Message) == 0) {	// Checksumme korrekt ?
			if (empfDaten.toAddr == myAddr) {				// bin ich gemeint?
				return true;
			}
		}
	}		
	return false;				// noch nicht fertig oder irgendwas ging schief
	
}


void handle_Serial_0_to_Serial_0 ()			
{	
	Led4_ON;								// fr Datalogger, debuggen
	if( read_Ringbuffer_0() == true)  {
		sendDaten.cmd = empfDaten.cmd;
		
		switch (empfDaten.cmd) {
			case   1:	sendDaten.data = TCNT1;						break;
			case   2:	sendDaten.data = OCR1A;						break;
			case   3:	sendDaten.data = OCR1B;						break;
			case   4:	sendDaten.data = countErrorsUART0;			break;	
			default:	sendDaten.data = last_MasterRequestTime;	
                  break;
		}
		
		send_Nachricht();
		last_MasterRequestTime = global_ms/1000;			// Zeitstempel letzter Kommunikation merken
		
	}
	Led4_OFF;
}


void send_Nachricht ()
{
	uint8_t data = 0;
	sendDaten.crc = calc_CRC16(sendDaten.asArray, (length_Message-sizeof(sendDaten.crc)) ) ;
	Led3_ON;						// fr Datalogger, debuggen
	MAX487_senden;					// MAX487 /RE_DE - umschalten auf senden
	
	uart0_putc(STX);
	for (uint8_t i=0; i < length_Message; i++) {
		data = sendDaten.asArray[i];
		if (data == ESC || data == ETX || data == STX) {
			uart0_putc(ESC);
		}
		uart0_putc(data);			// Bytes rausschieben
	}
	uart0_putc(ETX);
	
	while (UART0_TxHead != UART0_TxTail);	// warte auf leeren TX-Ringbuffer
	UCSR0A |= (1<<TXC0);					// lsche mit '1' "TXC0 UART Transmit Complete Flag"
	while( !(UCSR0A & (1<<TXC0)) );			// uart0_flush, warten bis Flag "TXC0 UART Transmit Complete" set
	
	MAX487_empfangen;						// MAX487 /RE_DE - umschalten auf empfangen
	Led3_OFF;
}


uint16_t calc_CRC16 (uint8_t feld[], uint8_t length)
{
	uint16_t crc = 0xFFFF;				// initial CRC value
	uint8_t b = 0;
	
	while (length--) {
		crc = _crc16_update(crc, feld[b++]);
	}
	return crc;
}


ISR(TIMER1_COMPB_vect)    // dauert 6,2s, wird aller >1ms aufgerufen
{	
	static uint8_t servo = 0;
	const uint8_t SERVO_CNT = 3;        // Anzahl Servos
	const uint8_t TIMING_WINDOW = 20;   // 1 Count = 1s
	static uint16_t temp = 0;
    
	switch(servo) {		// ISR Zeit 4,75s
		// Count   999 =  1ms
		// Count  1999 =  2ms
		// Count 19999 = 20ms
		case 0: Led2_OFF;	Led5_ON;	temp =  1499; break;
      case 1: Led5_OFF;	Led6_ON;	temp =  1499; break; 
		case 2: Led6_OFF;		      temp = 13999; break; // virtuelle
		case 3:           Led2_ON;	temp =  1499; break;
	}
    
	OCR1B += temp;
	OCR1A = OCR1B-TIMING_WINDOW;
   UCSR0B |= ((1<<RXCIE0) | (1<<UDRIE0) | (1<<TXB80));
	                                       // enable USART0 RX Complete Interrupt
											         // enable USART0 Data Register Empty Interrupt
   TCCR0B |= (1<<CS02)|(1<<CS00);	//run_Timer0();
   
	servo++;
	if (servo > SERVO_CNT)  servo = 0;
											
	Led1_OFF;					
}


ISR(TIMER1_COMPA_vect)			// dauert 1,8s
{	
   TCCR0B &= ~( (1<<CS02)|(1<<CS00) ); //stop_Timer0();
   
   UCSR0B &= ~((1<<RXCIE0) | (1<<UDRIE0) | (1<<TXB80));
	                              // disable USART0 RX Complete Interrupt
											// disable USART0 Data Register Empty Interrupt
	Led1_ON; // fr Datalogger, debuggen
}


